"""
cf_wms_router

OGC WMS router for datasets with CF convention metadata
"""
import io
import logging
import xml.etree.ElementTree as ET
from cmath import isnan
from typing import List

import cachey
import cf_xarray  # noqa
import numpy as np
import xarray as xr
from fastapi import Depends, HTTPException, Request, Response
from matplotlib import cm
from PIL import Image
from xpublish.dependencies import get_cache, get_dataset

from xpublish_wms.getmap import OgcWmsGetMap
from xpublish_wms.utils import (
    format_timestamp,
    lower_case_keys,
    round_float_values,
    speed_and_dir_for_uv,
    strip_float,
)

logger = logging.getLogger("uvicorn")


# WMS Styles declaration
# TODO: Add others beyond just simple raster
styles = [
    {
        "name": "raster/default",
        "title": "Raster",
        "abstract": "The default raster styling, scaled to the given range. The palette can be overridden by replacing default with a matplotlib colormap name",
    },
]


def create_text_element(root, name: str, text: str):
    element = ET.SubElement(root, name)
    element.text = text
    return element


def create_capability_element(root, name: str, url: str, formats: List[str]):
    cap = ET.SubElement(root, name)
    # TODO: Add more image formats
    for fmt in formats:
        create_text_element(cap, "Format", fmt)

    dcp_type = ET.SubElement(cap, "DCPType")
    http = ET.SubElement(dcp_type, "HTTP")
    get = ET.SubElement(http, "Get")
    get.append(
        ET.Element(
            "OnlineResource",
            attrib={"xlink:type": "simple", "xlink:href": url},
        ),
    )
    return cap


def create_parameter_feature_data(
    parameter,
    ds: xr.Dataset,
    has_time_axis,
    t_axis,
    x_axis,
    y_axis,
    values=None,
    name=None,
    id=None,
):
    # TODO Use standard and long name?
    name = (
        name
        if name is not None
        else ds[parameter].cf.attrs.get(
            "long_name",
            ds[parameter].cf.attrs.get("name", parameter),
        )
    )
    id = (
        id if id is not None else ds[parameter].cf.attrs.get("standard_name", parameter)
    )

    info = {
        "type": "Parameter",
        "description": {
            "en": name,
        },
        "observedProperty": {
            "label": {
                "en": name,
            },
            "id": id,
        },
    }

    axis_names = ["t", "x", "y"] if has_time_axis else ["x", "y"]
    shape = (
        [len(t_axis), len(x_axis), len(y_axis)]
        if has_time_axis
        else [len(x_axis), len(y_axis)]
    )
    values = values if values is not None else ds[parameter]
    values = round_float_values(values.squeeze().values.tolist())

    if isinstance(values, float):
        values = [values]

    range = {
        "type": "NdArray",
        "dataType": "float",
        # TODO: Some fields might not have a time field, and some might have an elevation field
        "axisNames": axis_names,
        "shape": shape,
        "values": [None if np.isnan(v) else v for v in values],
    }

    return (info, range)


def get_capabilities(ds: xr.Dataset, request: Request):
    """
    Return the WMS capabilities for the dataset
    """
    wms_url = f'{request.base_url}{request.url.path.removeprefix("/")}'

    root = ET.Element(
        "WMS_Capabilities",
        version="1.3.0",
        attrib={
            "xmlns": "http://www.opengis.net/wms",
            "xmlns:xlink": "http://www.w3.org/1999/xlink",
        },
    )

    service = ET.SubElement(root, "Service")
    create_text_element(service, "Name", "WMS")
    create_text_element(service, "Title", "XPublish WMS")
    create_text_element(service, "Abstract", "XPublish WMS")
    service.append(ET.Element("KeywordList"))
    service.append(
        ET.Element(
            "OnlineResource",
            attrib={
                "xlink:type": "simple",
                "xlink:href": "http://www.opengis.net/spec/wms_schema_1/1.3.0",
            },
        ),
    )

    capability = ET.SubElement(root, "Capability")
    # request_tag = ET.SubElement(capability, "Request")

    # get_capabilities = create_capability_element(
    #     request_tag,
    #     "GetCapabilities",
    #     wms_url,
    #     ["text/xml"],
    # )
    # TODO: Add more image formats
    # get_map = create_capability_element(request_tag, "GetMap", wms_url, ["image/png"])
    # TODO: Add more feature info formats
    # get_feature_info = create_capability_element(
    #     request_tag,
    #     "GetFeatureInfo",
    #     wms_url,
    #     ["text/json"],
    # )
    # TODO: Add more image formats
    # get_legend_graphic = create_capability_element(
    #     request_tag,
    #     "GetLegendGraphic",
    #     wms_url,
    #     ["image/png"],
    # )

    exeption_tag = ET.SubElement(capability, "Exception")
    exception_format = ET.SubElement(exeption_tag, "Format")
    exception_format.text = "text/json"

    layer_tag = ET.SubElement(capability, "Layer")
    create_text_element(layer_tag, "Title", ds.attrs.get("title", "Untitled"))
    create_text_element(
        layer_tag,
        "Description",
        ds.attrs.get("description", "No Description"),
    )
    create_text_element(layer_tag, "CRS", "EPSG:4326")
    create_text_element(layer_tag, "CRS", "EPSG:3857")
    create_text_element(layer_tag, "CRS", "CRS:84")

    for var in ds.data_vars:
        da = ds[var]

        # If there are not spatial coords, we can't view it with this router, sorry
        if "longitude" not in da.cf.coords:
            continue

        # bounds = {
        #     "CRS": "EPSG:4326",
        #     "minx": f'{da.cf.coords["longitude"].min().values.item()}',
        #     "miny": f'{da.cf.coords["latitude"].min().values.item()}',
        #     "maxx": f'{da.cf.coords["longitude"].max().values.item()}',
        #     "maxy": f'{da.cf.coords["latitude"].max().values.item()}',
        # }

        attrs = da.cf.attrs
        layer = ET.SubElement(layer_tag, "Layer", attrib={"queryable": "1"})
        create_text_element(layer, "Name", var)
        create_text_element(
            layer,
            "Title",
            attrs.get("long_name", attrs.get("name", var)),
        )
        create_text_element(
            layer,
            "Abstract",
            attrs.get("long_name", attrs.get("name", var)),
        )
        create_text_element(layer, "CRS", "EPSG:4326")
        create_text_element(layer, "CRS", "EPSG:3857")
        create_text_element(layer, "CRS", "CRS:84")

        create_text_element(layer, "Units", attrs.get("units", ""))

        # min_value = float(da.min())
        # create_text_element(layer, 'MinMag', min_value)

        # max_value = float(da.max())
        # create_text_element(layer, 'MaxMag', max_value)

        # Not sure if this can be copied, its possible variables have different extents within
        # a given dataset probably, but for now...
        # bounding_box_element = ET.SubElement(layer, "BoundingBox", attrib=bounds)

        if "T" in da.cf.axes:
            times = format_timestamp(da.cf["T"])

            time_dimension_element = ET.SubElement(
                layer,
                "Dimension",
                attrib={
                    "name": "time",
                    "units": "ISO8601",
                    "default": times[-1],
                },
            )
            # TODO: Add ISO duration specifier
            time_dimension_element.text = f"{','.join(times)}"

        style_tag = ET.SubElement(layer, "Style")

        for style in styles:
            style_element = ET.SubElement(
                style_tag,
                "Style",
                attrib={"name": style["name"]},
            )
            create_text_element(style_element, "Title", style["title"])
            create_text_element(style_element, "Abstract", style["abstract"])

            legend_url = f'{wms_url}?service=WMS&request=GetLegendGraphic&format=image/png&width=20&height=20&layers={var}&styles={style["name"]}'
            create_text_element(style_element, "LegendURL", legend_url)

    ET.indent(root, space="\t", level=0)
    get_caps_xml = ET.tostring(root).decode("utf-8")

    return Response(get_caps_xml, media_type="text/xml")


def get_feature_info(dataset: xr.Dataset, query: dict):
    """
    Return the WMS feature info for the dataset and given parameters
    """
    if not dataset.rio.crs:
        dataset = dataset.rio.write_crs(4326)

    if ":" in query["query_layers"]:
        parameters = query["query_layers"].split(":")
    else:
        parameters = query["query_layers"].split(",")
    times = list(dict.fromkeys([t.replace("Z", "") for t in query["time"].split("/")]))
    crs = query.get("crs", None) or query.get("srs")
    bbox = [float(x) for x in query["bbox"].split(",")]
    width = int(query["width"])
    height = int(query["height"])
    x = int(query["x"])
    y = int(query["y"])
    # format = query["info_format"]

    # We only care about the requested subset
    ds = dataset[parameters]

    # TODO: Need to reproject??

    x_coord = np.linspace(bbox[0], bbox[2], width)
    y_coord = np.linspace(bbox[1], bbox[3], height)

    has_time_axis = [
        ds[parameter].cf.axes.get("T") is not None for parameter in parameters
    ]
    any_has_time_axis = True in has_time_axis

    if any_has_time_axis:
        if len(times) == 1:
            resampled_data = ds.cf.interp(
                T=times[0],
                longitude=x_coord,
                latitude=y_coord,
            )
        elif len(times) > 1:
            resampled_data = ds.cf.interp(longitude=x_coord, latitude=y_coord)
            resampled_data = resampled_data.cf.sel(T=slice(times[0], times[1]))
        else:
            raise HTTPException(500, f"Invalid time requested: {times}")
    else:
        resampled_data = ds.cf.interp(longitude=x_coord, latitude=y_coord)

    x_axis = [strip_float(resampled_data.cf["longitude"][x])]
    y_axis = [strip_float(resampled_data.cf["latitude"][y])]
    resampled_data = resampled_data.cf.sel({"longitude": x_axis, "latitude": y_axis})

    # When none of the parameters have data, drop it
    if any_has_time_axis and resampled_data[resampled_data.cf.axes["T"][0]].shape:
        resampled_data = resampled_data.dropna(
            resampled_data.cf.axes["T"][0],
            how="all",
        )

    if not any_has_time_axis:
        t_axis = None
    elif len(times) == 1:
        t_axis = str(format_timestamp(resampled_data.cf["T"]))
    else:
        t_axis = str(format_timestamp(resampled_data.cf["T"]))

    parameter_info = {}
    ranges = {}

    for i_parameter, parameter in enumerate(parameters):
        info, range = create_parameter_feature_data(
            parameter,
            resampled_data,
            has_time_axis[i_parameter],
            t_axis,
            x_axis,
            y_axis,
        )
        parameter_info[parameter] = info
        ranges[parameter] = range

    # For now, hardcoding uv parameter grouping
    if len(parameters) == 2 and (
        "u_eastward" in parameters or "u_eastward_max" in parameters
    ):
        speed, direction = speed_and_dir_for_uv(
            resampled_data[parameters[0]],
            resampled_data[parameters[1]],
        )
        speed_info, speed_range = create_parameter_feature_data(
            parameter,
            resampled_data,
            has_time_axis[i_parameter],
            t_axis,
            x_axis,
            y_axis,
            speed,
            "Magnitude of velocity",
            "magnitude_of_velocity",
        )
        speed_parameter_name = f"{parameters[0]}:{parameters[1]}-mag"
        parameter_info[speed_parameter_name] = speed_info
        ranges[speed_parameter_name] = speed_range

        direction_info, direction_range = create_parameter_feature_data(
            parameter,
            resampled_data,
            has_time_axis[i_parameter],
            t_axis,
            x_axis,
            y_axis,
            direction,
            "Direction of velocity",
            "direction_of_velocity",
        )
        direction_parameter_name = f"{parameters[0]}:{parameters[1]}-dir"
        parameter_info[direction_parameter_name] = direction_info
        ranges[direction_parameter_name] = direction_range

    axis = (
        {"t": {"values": t_axis}, "x": {"values": x_axis}, "y": {"values": y_axis}}
        if any_has_time_axis
        else {"x": {"values": x_axis}, "y": {"values": y_axis}}
    )

    referencing = (
        [
            {
                "coordinates": ["t"],
                "system": {
                    "type": "TemporalRS",
                    "calendar": "gregorian",
                },
            },
            {
                "coordinates": ["x", "y"],
                "system": {
                    "type": "GeographicCRS",
                    "id": crs,
                },
            },
        ]
        if any_has_time_axis
        else [
            {
                "coordinates": ["x", "y"],
                "system": {
                    "type": "GeographicCRS",
                    "id": crs,
                },
            },
        ]
    )

    return {
        "type": "Coverage",
        "title": {
            "en": "Extracted Profile Feature",
        },
        "domain": {
            "type": "Domain",
            "domainType": "PointSeries",
            "axes": axis,
            "referencing": referencing,
        },
        "parameters": parameter_info,
        "ranges": ranges,
    }


def get_legend_info(dataset: xr.Dataset, query: dict):
    """
    Return the WMS legend graphic for the dataset and given parameters
    """
    parameter = query["layers"]
    width: int = int(query["width"])
    height: int = int(query["height"])
    vertical = query.get("vertical", "false") == "true"
    # colorbaronly = query.get("colorbaronly", "False") == "True"
    colorscalerange = [
        float(x) for x in query.get("colorscalerange", "nan,nan").split(",")
    ]
    if isnan(colorscalerange[0]):
        autoscale = True
    else:
        autoscale = query.get("autoscale", "false") != "false"
    style = query["styles"]
    stylename, palettename = style.split("/")

    ds = dataset.squeeze()

    # if the user has supplied a color range, use it. Otherwise autoscale
    if autoscale:
        min_value = float(ds[parameter].min())
        max_value = float(ds[parameter].max())
    else:
        min_value = colorscalerange[0]
        max_value = colorscalerange[1]

    scaled = (np.linspace(min_value, max_value, width) - min_value) / (
        max_value - min_value
    )
    data = np.ones((height, width)) * scaled

    if vertical:
        data = np.flipud(data.T)
        data = data.reshape((height, width))

    # Let user pick cm from here https://predictablynoisy.com/matplotlib/gallery/color/colormap_reference.html#sphx-glr-gallery-color-colormap-reference-py
    # Otherwise default to rainbow
    if palettename == "default":
        palettename = "rainbow"
    im = Image.fromarray(np.uint8(cm.get_cmap(palettename)(data) * 255))

    image_bytes = io.BytesIO()
    im.save(image_bytes, format="PNG")
    image_bytes = image_bytes.getvalue()

    return Response(content=image_bytes, media_type="image/png")


def wms_root(
    request: Request,
    dataset: xr.Dataset = Depends(get_dataset),
    cache: cachey.Cache = Depends(get_cache),
):
    query_params = lower_case_keys(request.query_params)
    method = query_params.get("request", "").lower()
    logger.info(f"WMS: {method}")
    if method == "getcapabilities":
        return get_capabilities(dataset, request)
    elif method == "getmap":
        getmap_service = OgcWmsGetMap()
        getmap_service.cache = cache
        return getmap_service.get_map(dataset, query_params)
    elif method == "getfeatureinfo" or method == "gettimeseries":
        return get_feature_info(dataset, query_params)
    elif method == "getlegendgraphic":
        return get_legend_info(dataset, query_params)
    else:
        raise HTTPException(
            status_code=404,
            detail=f"{method} is not a valid option for REQUEST",
        )



import io
import logging
import time
from datetime import datetime
from typing import List

import cachey
import cf_xarray  # noqa
import numpy as np
import pandas as pd
import xarray as xr
from fastapi.responses import StreamingResponse
from matplotlib import cm
from PIL import Image
from pykdtree.kdtree import KDTree
from rasterio.enums import Resampling
from rasterio.transform import from_bounds

from xpublish_wms.utils import lnglat_to_cartesian, to_lnglat

logger = logging.getLogger(__name__)


class OgcWmsGetMap:
    TIME_CF_NAME: str = "time"
    DEFAULT_CRS: str = "4326"
    DEFAULT_STYLE: str = "raster/default"

    cache: cachey.Cache

    # Data selection
    parameter: str
    time: datetime = None

    # Grid
    crs: str
    bbox = List[float]
    width: int
    height: int

    # Output style
    style: str
    colorscalerange: List[float]
    colorbaronly: bool
    autoscale: bool

    def get_map(self, ds: xr.Dataset, query: dict):
        """
        Return the WMS map for the dataset and given parameters
        """
        # Decode request params
        self.ensure_query_types(query)

        # Select data according to request
        da = self.select_layer(ds)
        da = self.select_time(da)
        da = self.select_custom_dim(da)
        da_bbox = self.select_grid(da)

        # Generate output
        image_bytes = self.draw(da, da_bbox)

        return StreamingResponse(image_bytes, media_type="image/png")

    def ensure_query_types(self, query: dict):
        """
        Decode request params

        :param query:
        :return:
        """
        # Data selection
        self.parameter = query["layers"]
        time_str = query.get("time", None)
        if time_str:
            self.time = pd.to_datetime(time_str).tz_localize(None)

        # Grid
        self.crs = query.get("crs", None) or query.get("srs")
        self.bbox = [float(x) for x in query["bbox"].split(",")]
        self.width = int(query["width"])
        self.height = int(query["height"])

        # Output style
        self.style = query.get("styles", self.DEFAULT_STYLE)
        self.colorscalerange = [
            float(x) for x in query.get("colorscalerange", "nan,nan").split(",")
        ]
        self.autoscale = query.get("autoscale", "true") == "true"

    def select_layer(self, ds: xr.Dataset) -> xr.DataArray:
        """
        Select Dataset variable, according to WMS layers request
        :param ds:
        :return:
        """
        return ds[self.parameter]

    def select_time(self, da: xr.DataArray) -> xr.DataArray:
        """
        Ensure time selection

        If time is provided :
            - use cf_xarray to access time
            - by default use TIME_CF_NAME
            - method nearest to ensure at least one result

        Otherwise:
            - Get latest one

        :param da:
        :return:
        """
        if self.time is not None:
            da = da.cf.sel({self.TIME_CF_NAME: self.time}, method="nearest")
        else:
            da = da.cf.isel({self.TIME_CF_NAME: -1})

        return da

    def select_custom_dim(self, da: xr.DataArray) -> xr.DataArray:
        """
        Select other dimension, ensuring a 2D array
        :param da:
        :return:
        """
        # Squeeze single value dimensions
        da = da.squeeze()

        # TODO: Filter dimension from custom query, if any

        # Squeeze multiple values dimensions, by selecting the last value
        for coord_name in da.cf.coords:
            if coord_name in ("latitude", "longitude", "X", "Y"):
                continue
            coord = da.cf.coords[coord_name]
            if coord.size > 1:
                da = da.cf.isel({coord_name: -1})

        return da

    def select_grid(self, da: xr.DataArray) -> xr.DataArray:
        """
        Select grid and reproject if needed
        :param da:
        :return:
        """
        # Some basic check for dataset
        if not da.rio.crs:
            da = da.rio.write_crs(self.DEFAULT_CRS)

        if self.grid_is_regular(da):
            da = self.select_regular_grid(da)
        else:
            da = self.select_irregular_grid(da)
        return da

    def grid_is_regular(self, da: xr.DataArray) -> bool:
        return da.cf.coords["longitude"].dims[0] == da.cf.coords["longitude"].name

    def select_regular_grid(self, da: xr.DataArray) -> xr.DataArray:
        """
        Filter regular grid according to WMS request :
            - bbox
            - width and height
        :param da:
        :return:
        """
        minx, miny, maxx, maxy = self.bbox

        transform = from_bounds(
            west=minx,
            south=miny,
            east=maxx,
            north=maxy,
            width=self.width,
            height=self.height,
        )
        clipped = da.rio.clip_box(
            minx=minx,
            miny=miny,
            maxx=maxx,
            maxy=maxy,
            crs=self.crs,
        )
        resampled_data = clipped.rio.reproject(
            dst_crs=self.crs,
            shape=(self.width, self.height),
            resampling=Resampling.nearest,
            transform=transform,
        )

        return resampled_data

    def select_irregular_grid(self, da: xr.DataArray) -> xr.DataArray:
        """
        Filter irregular grid according to WMS request :
            - bbox
            - width and height
        :param da:
        :return:
        """
        ds = da
        bbox, width, height = self.bbox, self.width, self.height

        start = time.time()
        min_lng = ds.cf.coords["longitude"].min().values.item()
        min_lat = ds.cf.coords["latitude"].min().values.item()
        max_lng = ds.cf.coords["longitude"].max().values.item()
        max_lat = ds.cf.coords["latitude"].max().values.item()

        # Check if we need to project the bounding box
        if self.crs == "EPSG:3857":
            t_lng, t_lat = to_lnglat.transform([bbox[0], bbox[2]], [bbox[1], bbox[3]])
        else:
            t_lng = [bbox[0], bbox[2]]
            t_lat = [bbox[1], bbox[3]]

        lngs = np.linspace(t_lng[0], t_lng[1], width)
        lats = np.linspace(t_lat[0], t_lat[1], height)

        grid_lngs, grid_lats = np.meshgrid(lngs, lats)

        pts = lnglat_to_cartesian(grid_lngs.ravel(), grid_lats.ravel())

        # Need ll version for masking outside dataset bounds
        pts_ll = np.column_stack((grid_lngs.ravel(), grid_lats.ravel()))
        pts_ll_mask = np.array(
            [
                x[0] >= min_lng
                and x[0] <= max_lng
                and x[1] >= min_lat
                and x[1] <= max_lat
                for x in pts_ll
            ],
        )

        if np.any(pts_ll_mask):
            kd = get_spatial_kdtree(ds, self.cache)
            dist, n = kd.query(pts)

            d_lng = pts[1][0] - pts[0][0]
            d_lat = pts[1][1] - pts[0][1]
            d_ele = pts[1][2] - pts[0][2]
            max_dist = np.sqrt((2 * d_lng) ** 2 + (2 * d_lat) ** 2 + (2 * d_ele))
            dist_mask = np.where(dist > max_dist)

            logger.info(f"Calculated max dist: {max_dist}")
            logger.info(f"max dist: {np.max(dist)}")
            logger.info(f"min dist: {np.min(dist)}")
            logger.info(f"mean dist: {np.mean(dist)}")
            logger.info(f"median dist: {np.median(dist)}")
            logger.info(f"stdev dist: {np.std(dist)}")
            logger.info("-----------------")

            ni = n.argsort()
            pp = n[ni]

            index_time = time.time()
            logger.info(f"index and kdtree irregular: {index_time - start}")

            # This is slow because it has to pull into numpy array, can we do better?
            # TODO: Can we avoid pulling down fully masked chunks???
            z = ds.zeta[0][pp].values
            z = z[ni.argsort()]
            z[~pts_ll_mask] = np.nan
            z[dist_mask] = np.nan

            z = z.reshape((height, width))

            extraction_time = time.time()
            logger.info(f"extract data irregular: {extraction_time - index_time}")

            rds = xr.Dataset(
                data_vars=dict(
                    z=(["y", "x"], z),
                ),
                coords=dict(
                    x=(["x"], lngs),
                    y=(["y"], lats),
                ),
            )
            rds.rio.write_crs(4326, inplace=True)
            resampled_data = rds.z.rio.reproject(
                dst_crs=self.crs,
                shape=(width, height),
                resampling=Resampling.nearest,
                transform=from_bounds(*bbox, width=width, height=height),
            )

            reproject_time = time.time()
            logger.info(
                f"clip and reproject irregular: {reproject_time - extraction_time}",
            )
        else:
            resampled_data = np.empty((width, height))
            resampled_data[:] = np.nan

        reproject_time = time.time()
        logger.info(f"clip and reproject irregular: {reproject_time - extraction_time}")
        return resampled_data

    def draw(self, da: xr.DataArray, da_bbox: xr.DataArray) -> io.BytesIO:
        """
        Generate drawing, could be easily overridden

        :param da:
        :param da_bbox:
        :return:
        """
        da_scaled = self.draw_pil_get_colormap_scaled_data(da, da_bbox)
        return self.draw_pil_generate_map(da_scaled)

    def draw_pil_get_colormap_scaled_data(
        self,
        da: xr.DataArray,
        da_bbox: xr.DataArray,
    ) -> xr.DataArray:
        """
        Generate numpy array from our datasset, ensuring colormap is computed from the
        non-clipped data
        :param da:
        :param da_bbox:
        :return:
        """
        if self.autoscale:
            min_value = float(da.min())
            max_value = float(da.max())
        else:
            min_value = self.colorscalerange[0]
            max_value = self.colorscalerange[1]

        return (da_bbox - min_value) / (max_value - min_value)

    def draw_pil_generate_map(self, da: xr.DataArray) -> io.BytesIO:
        """
        Draw as PIL.Image
        :param da:
        :return:
        """
        try:
            stylename, palettename = self.style.split("/")
        except:  # noqa: E722
            palettename = "default"

        # Let user pick cm from here https://predictablynoisy.com/matplotlib/gallery/color/colormap_reference.html#sphx-glr-gallery-color-colormap-reference-py
        # Otherwise default to rainbow
        if palettename == "default":
            palettename = "rainbow"
        im = Image.fromarray(np.uint8(cm.get_cmap(palettename)(da) * 255))

        image_bytes = io.BytesIO()
        im.save(image_bytes, format="PNG")
        image_bytes.seek(0)

        return image_bytes


def get_spatial_kdtree(ds: xr.Dataset, cache: cachey.Cache) -> KDTree:
    cache_key = f"dataset-kdtree-{ds.attrs['title']}"
    kd = cache.get(cache_key)
    if kd:
        return kd

    lng = ds.cf["longitude"]
    lat = ds.cf["latitude"]

    verts = lnglat_to_cartesian(lng, lat)
    kd = KDTree(verts)

    cache.put(cache_key, kd, 5)

    return kd



"""
xpublish_wms is not a real package, just a set of best practices examples.
"""

from .plugin import CfWmsPlugin

__all__ = ["CfWmsPlugin"]

try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"



import logging
from typing import Union

import numpy as np
import xarray as xr
from pyproj import Transformer

logger = logging.getLogger(__name__)


def lower_case_keys(d: dict) -> dict:
    return {k.lower(): v for k, v in d.items()}


def format_timestamp(value):
    return value.dt.strftime(date_format="%Y-%m-%dT%H:%M:%SZ").values


def strip_float(value):
    return float(value.values)


def round_float_values(v: list) -> list:
    if not isinstance(v, list):
        return round(v, 5)
    return [round(x, 5) for x in v]


def speed_and_dir_for_uv(u, v):
    """
    Given u and v values or arrays, calculate speed and direction transformations
    """
    speed = np.sqrt(u**2 + v**2)

    dir_trig_to = np.arctan2(u / speed, v / speed)
    dir_trig_deg = dir_trig_to * 180 / np.pi
    dir = (dir_trig_deg) % 360

    return [speed, dir]


def ensure_crs(
    ds: Union[xr.Dataset, xr.DataArray],
    default_crs: str = "EPSG:4326",
) -> Union[xr.Dataset, xr.DataArray]:
    """
    Ensure our dataset has a CRS
    :param ds:
    :param default_crs:
    :return:
    """
    # logger.debug(f"CRS found in dataset : {ds.rio.crs}")
    if not ds.rio.crs:
        logger.debug(f"Settings default CRS : {default_crs}")
        ds.rio.write_crs(default_crs, inplace=True)
    return ds


def lnglat_to_cartesian(longitude, latitude):
    """
    Converts latitude and longitude to cartesian coordinates
    """
    lng_rad = np.deg2rad(longitude)
    lat_rad = np.deg2rad(latitude)

    R = 6371
    x = R * np.cos(lat_rad) * np.cos(lng_rad)
    y = R * np.cos(lat_rad) * np.sin(lng_rad)
    z = R * np.sin(lat_rad)
    return np.column_stack((x, y, z))


to_lnglat = Transformer.from_crs(3857, 4326, always_xy=True)



import logging
from typing import List

import cachey
import xarray as xr
from fastapi import APIRouter, Depends, Request
from xpublish import Dependencies, Plugin, hookimpl

import xpublish_wms.cf_wms as cf_wms

logger = logging.getLogger("cf_wms")


class CfWmsPlugin(Plugin):
    """
    OGC WMS plugin for xpublish
    """

    name = "cf_wms"

    dataset_router_prefix: str = "/wms"
    dataset_router_tags: List[str] = ["wms"]

    @hookimpl
    def dataset_router(self, deps: Dependencies) -> APIRouter:
        """Register dataset level router for WMS endpoints"""

        router = APIRouter(
            prefix=self.dataset_router_prefix,
            tags=self.dataset_router_tags,
        )

        @router.get("/")
        def wms_root(
            request: Request,
            dataset: xr.Dataset = Depends(deps.dataset),
            cache: cachey.Cache = Depends(deps.cache),
        ):
            return cf_wms.wms_root(request, dataset, cache)

        return router
